(function() {
    // define namespace
    window.SPCORS = window.SPCORS || {};
    
    // helper returns true if str matches any RegExp in arr
    var matches = function(str, arr) {
    	return !arr || (str && arr.reduce(function(acc, url) {
          return acc || url.test(str);
        }, false));
    };
    
    // helper throws exception if str does not match any RegExp in arr
    var gaurde = function(str, arr, type) {
    	if(!matches(str, arr)) {
        	throw new Error(type + ": " + str + 
            	" does not match [" + arr.join() + "].");
        }
    };
    
    // helper converts a url to an origin
    var baseUrl = function(url) {
        var parts = url.split("/");
        return parts.length > 3 ? parts.splice(0, 3).join("/") : url;
    };
    
    /**
     * Represents a proxy that listens for postMessage events and attempts to
     * convert them to SharePoint ajax requests using fetch, send the reponse back as a 
     * postMessage to the parent window (assumes running in an iframe).
     * @constructor
     * @param (string) origin - The url expected for the response origin, will
     * accept any origin if set to "*".
     * @param [string] originPatterns - message handler will post an exception
     * message if this array is defined and contains at least one RegExp and the 
     * origin does not match any of these patterns.
     * @param [string] urlPatterns - ajax will post an exception message if this
     * array is defined and contains at least one RegExp and the 
     * ajax url does not match any of these patterns.
     * @returns {object} a new ServiceProxy instance.
     */
    function ServiceProxy(origin, originPatterns, urlPatterns) {
    	var _self = this;
        
        /**
         * Call fetch and send a post message with either the result or an error
         * message.
         * @param {object} init - the input to fetch. Note that fetch takes the url as a 
         * separate parameter, but I do not, because post message can only easily
         * send a single object, so I split out the url before calling fetch.
         * @returns undefined - the result is sent back to the parent window as a post
         * message using the respond method.
         */
        this.fetchWrapper = function(init) {
            init = init || {};
            var url = init.url; // split out the url
            
            try {
                // transform substitution tokens
                url = url.replace(/~site/g, _spPageContextInfo.webAbsoluteUrl);
                url = url.replace(/~sitecollection/g, _spPageContextInfo.siteAbsoluteUrl);
                
                // throw exception if urlPatterns test fails
                gaurde(url, urlPatterns, "URL");
                
                // always include credentials
                init.credentials = "include";

                // turn arbitrary objects into Headers
                if (init && init.headers && !(init.headers instanceof Headers)) {
                    var newHeaders = new Headers();
                    for (var key in init.headers) {
                        newHeaders.append(key, init.headers[key]);
                    }
                    init.headers = newHeaders;
                }
                
                // update the form digest if necessary
                UpdateFormDigest(_spPageContextInfo.webServerRelativeUrl, _spFormDigestRefreshInterval);

                // get the request digest, this has to be gotten in the service proxy 
                // since the digest is site collection specific and we're cross-origin
                init.headers["X-RequestDigest"] = document.getElementById("__REQUESTDIGEST").value;
                    
                if(init.body && init.body.match(/^data:[a-z]*\/[a-z]*;base64,/i)) {
                    this.dataUri2Blob(init);
                }
            }
            catch(e) {
                // send the error as a post message
                _self.respond(init, "exception", e.message + "\n\n" + e.stack);
            }
            
            // Call fetch just as you would the global fetch, but headers
            // can be an arbitrary JSON object.
            fetch(url, init)
                .then(function(response) {
                    // non-success response
                    if (response.status < 200 || response.status >= 400) {
                        throw new Error(response.status + " " + response.statusText);
                    }

                    // no content by design
                    if (response.status === 204) {
                        return response.text();
                    }

                    // no content, but not not by design?
                    if (response.headers.get("content-length") === "0") {
                        return response.text();
                    }

                    // process json responses
                    if(init.headers) {
                        var accept = init.headers.get("accept");
                        if (accept && accept.indexOf("application/json") > -1) {
                            var contentType = response.headers.get("content-type");
                            if (contentType && contentType.indexOf("application/json") > -1) {
                                return response.json();
                            } else {
                                throw new TypeError("Oops, should have gotten JSON, but didn't!");
                            }
                        }
                    }
                    
                    // no error, but also not JSON
                    return response.text();
                }).then(function(data) {
                    // send the response as a post message
                    _self.respond(init, "ok", data); 
                }).catch(function(error) {
                    // send the error as a post message
                    _self.respond(init, "exception", error.message + "\n\n" + error.stack);
                });
            
            return;
        }

 /*
Adapted from dataURItoBlob - https://gist.github.com/mcsheaj/ae358dc81d9061d74e1c986436052d4f
The MIT License (MIT)
Copyright (c) 2016 David Gomez-Urquiza
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
        /**
         * Convert a data url body to binary.
         * @init {object} - the fetch init object. This method assumes the body is a
         * data uri, so check first.
         * @returns undefined - the body of the init object is updated with the binary
         * contents if successful.
         */
        this.dataUri2Blob = function(init) {
            // get the base64 encoded file contents from the data url
            var bytes = atob(init.body.split(',')[1]);
            // get the mime type from the data url
            var mimeType = init.body.split(',')[0].split(':')[1].split(';')[0];
                
            // create an array buffer and decode the bytes to it
            var buffer = new ArrayBuffer(bytes.length);
            var uia = new Uint8Array(buffer);
            for (var i = 0; i < bytes.length; i++) {
                uia[i] = bytes.charCodeAt(i);
            }
                
            try {
                // create a data view and instantiate a new Blob by mime type
                var dataView = new DataView(buffer);
                init.body = new Blob([dataView], { type: mimeType });
            }
            catch (e) {
                // Old browser, need to use blob builder, including all versions of
                // internet explorer
                window.BlobBuilder = window.BlobBuilder ||
                    window.WebKitBlobBuilder ||
                    window.MozBlobBuilder ||
                    window.MSBlobBuilder;
                    
                if (window.BlobBuilder) {
                    // if there is a blob builder, append the array buffer to it
                    // and get the contents by mime type.
                    var blobBuilder = new BlobBuilder();
                    blobBuilder.append(buffer);
                    init.body = blobBuilder.getBlob(mimeType);
                }
            }
            
            return;
        },
        
        /**
         * Send an fetch response (or exception) back to the calling frame via
         * post message.
         * @param {options} - The options passed to fetch.
         * @param (string) - A status message included in the response, generally
         * either 'ok' or 'exception'.
         * @param {data} the data returned from the web service, or a string 
         * exception message.
         * @return (undefined) will call respond to send the response as a 
         * post message.
         */
        this.respond = function(options, status, data) {
         	var response = {
               	options: options,
                status: status,
                message: data
            };
            
            parent.postMessage(JSON.stringify(response), origin);
                
            return;
        };
        
        /**
         * Create a listener for postMessage events and pass the data to fetch
         */
        window.addEventListener("message", function(e) {
        	if(typeof(e.data) === "string") {
            	var message;
                try {
                    // deserialize the json data
            	    message = JSON.parse(e.data);
                } catch(e) { /* assume message for someone else */ }
                
                // if not json, then not a message for this handler
                if(message) {
                    // throw an exception if originPatterns do not match origin
                	gaurde(e.origin, originPatterns, "Origin");
                    // pass the message to this.ajax
                	_self.fetchWrapper(message);
                }
            }
        }, false);
    }
    // attach class definition to namespace
    SPCORS.ServiceProxy = ServiceProxy;
    
    /**
     * Represents a consumer of SharePoint ajax requests using postMessage to
     * execute requests safely in another domain (i.e. a consumer of postMessages
     * from ServiceProxy).
     * @constructor
     * @param (string) iframeid - the id of the iframe DOM element that implements
     * the ServiceProxy.
     * @param (string) origin - The url expected for the response origin, will
     * accept any origin if set to "*".
     * @param [string] originPatterns - message handler will post an exception
     * message if this array is defined and contains at least one RegExp and the 
     * origin does not match any of these patterns.
     * @returns {object} a new ProxyClient instance.
     */
    function ProxyClient(iframeid, origin, originPatterns) {
    	var _self = this;
        
        // Private map of promises by guid (the guid 
        // is generated when the request is made). When a post message is 
        // recieved, the message handler looks for the guid, retrieves the 
        // promise, resolves/rejects the promise, and finally deletes the promise from the map
        // so they it can only be called once and the map doesn't grow indefinitely.
        var promises = {};
        
        /**
         * Use this method just as you would use fetch.
         * @param {string} url - the url to use for the ajax request.
         * @param {object} init - the initialization object for fetch. See fetch
         * documentation for details.
         * @returns {object} - a promise which will be resolved or recjected based
         * on the outcome of the fetch operation.
         */
        this.fetch = function(url, init) {
            // create a new promise
            var promise = new Promise(function(resolve, reject) {
                // get a handle to the iframe
                var iframe = document.getElementById(iframeid).contentWindow;
                
                // add the url to the init object to be passed over
                // to the service proxy.
                init.url = url;
                
                // create a new guid for the request
                init.guid = SP.Guid.newGuid().toString();
                
                // add the promise by guid to the promises map, along
                // with it's resolve and reject callbacks, which will get
                // call back later by the post message handler
                promises[init.guid] = {
                    promise: promise,
                    resolve: resolve,
                    reject: reject
                };

                // post the options as a message to the proxy frame
                iframe.postMessage(JSON.stringify(init), origin);
            });
            
            // return the promise 
            return promise;
        };
        
        /*
         * Crate a listener for postMessages and pass the data to this.ajax.
         */
        window.addEventListener("message", function(e) {
            var promise;
            try {
                // we expect a string, if not the message may be for a 
                // different listener
                if(typeof(e.data) !== "string")
                    return;

                // deserialize the message
                var data = e.data.trim();
                try {
                    data = JSON.parse(data);
                } catch(e) {}

                // we expect an object, with an options property that is also
                // an object, if not the message may be for a different listener
                if(typeof(data) !== "object" || typeof(data.options) !== "object") 
                    return;

                // if we have a promise for this message for this message
                if(data.options.guid && promises[data.options.guid]) {
                    // get the promise
                    promise = promises[data.options.guid];

                    if(window !== window.top) {
                        // if the origin doesn't match expectations, throw an exception
                        gaurde(e.origin, originPatterns, "Origin");
                    }
                    
                    if(data.status === "exception") {
                        // exception, reject the promise
                        promise.reject(data.message);
                    }
                    else {
                        // resolve the promise
                        promise.resolve(data.message);                    
                    }
                }
            }
            catch(e) {
                // reject the promise if we got far enough to find one
                if(typeof(promise) !== "undefined") promise.reject(e);
            }
        }, false);
    }
    // attach class definition to namespace
    SPCORS.ProxyClient = ProxyClient;
})();